import logging
import time
from collections.abc import Callable, Generator, Iterable, Mapping
from concurrent import futures
from datetime import datetime
from json.decoder import JSONDecodeError

# TODO Use `type` instead of TypeAlias when dropping Python 3.11
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeAlias, TypeVar, cast

import httpx
import numpy as np
import pandas as pd
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, Field, GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4.conf import settings
from ixmp4.core.exceptions import (
    ApiEncumbered,
    ImproperlyConfigured,
    IxmpError,
    UnknownApiError,
    registry,
)
from ixmp4.data import abstract

if TYPE_CHECKING:
    from ixmp4.data.backend.api import RestBackend

logger = logging.getLogger(__name__)

JsonType: TypeAlias = Mapping[
    str,
    Iterable[float]
    | Iterable[int]
    | Iterable[str]
    | Mapping[str, Any]
    | abstract.annotations.PrimitiveTypes
    | None,
]
ParamType: TypeAlias = dict[
    str, bool | int | str | list[int] | Iterable[int] | Mapping[str, Any] | None
]
_RequestParamType: TypeAlias = Mapping[
    str,
    abstract.annotations.PrimitiveTypes
    | abstract.annotations.PrimitiveIterableTypes
    | Mapping[str, Any]
    | None,
]


class BaseModel(PydanticBaseModel):
    NotFound: ClassVar[type[IxmpError]]
    NotUnique: ClassVar[type[IxmpError]]

    model_config = ConfigDict(from_attributes=True)


class DataFrameDict(TypedDict):
    index: list[int] | list[str]
    columns: list[str]
    dtypes: list[str]
    # NOTE This is deliberately slightly out of sync with DataFrame.data below
    # (cf positioning of int and float) to demonstrate that only the one below seems to
    # affect our tests by causing ValidationErrors
    data: list[
        list[
            abstract.annotations.PrimitiveTypes
            | datetime
            | dict[str, Any]
            | list[float]
            | list[int]
            | list[str]
            | None
        ]
    ]


def df_to_dict(df: pd.DataFrame) -> DataFrameDict:
    columns = []
    dtypes = []
    for c in df.columns:
        columns.append(c)
        dtypes.append(df[c].dtype.name)

    return DataFrameDict(
        index=df.index.to_list(),
        columns=columns,
        dtypes=dtypes,
        data=df.values.tolist(),
    )


class DataFrame(PydanticBaseModel):
    index: list[int] | list[str] | None = Field(None)
    columns: list[str] | None
    dtypes: list[str] | None
    # TODO The order is important here at the moment, in particular having int before
    # float! This should likely not be the case, but using StrictInt and StrictFloat
    # from pydantic only created even more errors.
    data: (
        list[
            list[
                bool
                | datetime
                | int
                | float
                | str
                | dict[str, Any]
                | list[float]
                | list[int]
                | list[str]
                | None
            ]
        ]
        | None
    )

    model_config = ConfigDict(json_encoders={pd.Timestamp: lambda x: x.isoformat()})

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_before_validator_function(cls.validate, handler(cls))

    @classmethod
    def validate(cls, df: pd.DataFrame | DataFrameDict) -> DataFrameDict:
        return df_to_dict(df) if isinstance(df, pd.DataFrame) else df

    def to_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame(
            index=self.index or None,
            columns=self.columns,
            data=self.data,
        )
        if self.columns and self.dtypes:
            for c, dt in zip(self.columns, self.dtypes):
                # there seems to be a type incompatbility between StrDtypeArg and str
                df[c] = df[c].astype(dt)  # type: ignore[call-overload]
        return df


class _RequestKwargs(TypedDict, total=False):
    params: _RequestParamType | None
    json: JsonType | None
    max_retries: int


class RequestKwargs(TypedDict, total=False):
    content: str


ModelType = TypeVar("ModelType", bound=BaseModel)


class BaseRepository(Generic[ModelType]):
    model_class: type[ModelType]
    prefix: ClassVar[str]
    enumeration_method: str = "PATCH"

    backend: "RestBackend"

    def __init__(self, backend: "RestBackend") -> None:
        self.backend = backend

    def sanitize_params(self, params: Mapping[str, Any]) -> dict[str, Any]:
        return {k: params[k] for k in params if params[k] is not None}

    def get_remote_exception(self, res: httpx.Response, status_code: int) -> IxmpError:
        try:
            json = res.json()
        except (ValueError, JSONDecodeError):
            raise UnknownApiError(content=res.text, response_status=res.status_code)

        try:
            error_name = json["error_name"]
        except KeyError:
            raise UnknownApiError(json_data=json, response_status=status_code)

        try:
            exc = registry[error_name]
        except KeyError:
            raise ImproperlyConfigured(
                "Could not find remote exception in registry. "
                "Are you sure client and server ixmp versions are compatible?"
            )
        return exc.from_dict(json)

    def _request(
        self,
        method: str,
        path: str,
        params: _RequestParamType | None = None,
        json: JsonType | None = None,
        max_retries: int = settings.client_max_request_retries,
        **kwargs: Unpack[RequestKwargs],
    ) -> dict[str, Any] | list[Any] | None:
        """Sends a request and handles potential error responses.
        Re-raises a remote `IxmpError` if thrown and transferred from the backend.
        Handles read timeouts and rate limiting responses via retries with backoffs.
        Returns `None` if the response body is empty
        but has a status code less than 300.
        """

        def retry(max_retries: int = max_retries) -> dict[str, Any] | list[Any] | None:
            if max_retries == 0:
                logger.error(f"API Encumbered: '{self.backend.info.dsn}'")
                raise ApiEncumbered(
                    f"The service connected to the backend '{self.backend.info.name}' "
                    "is currently overencumbered with requests. "
                    "Try again at a later time or re-configure your client "
                    "to behave more leniently."
                )
            # increase backoff by `settings.client_backoff_factor`
            # for each retry
            backoff_s = settings.client_backoff_factor * (
                settings.client_max_request_retries - max_retries
            )
            logger.debug(f"Retrying request in {backoff_s} seconds.")
            time.sleep(backoff_s)
            return self._request(
                method,
                path,
                params=params,
                json=json,
                max_retries=max_retries - 1,
                **kwargs,
            )

        params = self.sanitize_params(params) if params else {}

        try:
            res = self.backend.client.request(
                method,
                path,
                params=params,
                json=json,
                **kwargs,
            )
        except httpx.ReadTimeout:
            logger.warning("Read timeout, retrying request...")
            return retry()

        return self._handle_response(res, retry)

    def _handle_response(
        self,
        res: httpx.Response,
        retry: Callable[..., dict[str, Any] | list[Any] | None],
    ) -> dict[str, Any] | list[Any] | None:
        if res.status_code in [
            429,  # Too Many Requests
            420,  # Enhance Your Calm
        ]:
            return retry()
        elif res.status_code >= 400:
            if res.status_code == 413:
                raise ImproperlyConfigured(
                    "Received status code 413 (Payload Too Large). "
                    "Consider decreasing `IXMP4_CLIENT_DEFAULT_UPLOAD_CHUNK_SIZE` "
                    f"(current: {settings.client_default_upload_chunk_size})."
                )
            raise self.get_remote_exception(res, res.status_code)
        else:
            try:
                # res.json just returns Any...
                json_decoded: dict[str, Any] | list[Any] = res.json()
                return json_decoded
            except JSONDecodeError:
                if res.status_code < 300 and res.text == "":
                    return None
            except ValueError:
                pass
            raise UnknownApiError(res.text)

    def _get_by_id(self, id: int) -> dict[str, Any]:
        # we can assume this type on create endpoints
        return self._request("GET", self.prefix + str(id) + "/")  # type: ignore[return-value]

    def _request_enumeration(
        self,
        table: bool = False,
        params: ParamType | None = None,
        json: JsonType | None = None,
        path: str = "",
    ) -> dict[str, Any] | list[Any]:
        """Convenience method for requests to the enumeration endpoint."""
        if params is None:
            params = {}

        # See https://github.com/iiasa/ixmp4/pull/129#discussion_r1841829519 for why we
        # are keeping this assumption
        # we can assume these types on enumeration endpoints
        return self._request(
            self.enumeration_method,
            self.prefix + path,
            params={**params, "table": table},
            json=json,
        )  # type: ignore[return-value]

    def _dispatch_pagination_requests(
        self,
        total: int,
        start: int,
        limit: int,
        params: ParamType | None,
        json: JsonType | None,
        path: str = "",
    ) -> list[list[Any]] | list[dict[str, Any]]:
        """Uses the backends executor to send many pagination requests concurrently."""
        requests: list[futures.Future[dict[str, Any]]] = []
        for req_offset in range(start, total, limit):
            req_params = params.copy() if params is not None else {}

            req_params.update({"limit": limit, "offset": req_offset})
            # Based on usage below, we seem to rely on self._request always returning a
            # dict[str, Any] here
            futu: futures.Future[dict[str, Any]] = self.backend.executor.submit(
                self._request,  # type: ignore [arg-type]
                self.enumeration_method,
                self.prefix + path,
                params=req_params,
                json=json,
            )
            requests.append(futu)
        results = futures.wait(requests)
        responses = [f.result() for f in results.done]
        # This seems to imply that type(responses) == list[dict[str, Any]]
        return [r.pop("results") for r in responses]

    def _handle_pagination(
        self,
        data: dict[str, Any],
        table: bool = False,
        params: ParamType | None = None,
        json: JsonType | None = None,
        path: str = "",
    ) -> list[list[Any]] | list[dict[str, Any]]:
        """Handles paginated response and sends subsequent requests if necessary.
        Returns aggregated pages as a list."""

        if params is None:
            params = {"table": table}
        else:
            params["table"] = table

        total = data.pop("total")
        pagination = data.pop("pagination")
        offset = pagination.pop("offset")
        limit = pagination.pop("limit")
        if total <= offset + limit:
            return [data.pop("results")]
        else:
            results = self._dispatch_pagination_requests(
                total, offset + limit, limit, params, json, path=path
            )
            return [data.pop("results")] + results

    def _list(
        self,
        params: ParamType | None = None,
        json: JsonType | None = None,
        path: str = "",
    ) -> list[ModelType]:
        data = self._request_enumeration(params=params, table=False, json=json)
        if isinstance(data, dict):
            # we can assume this type on list endpoints
            pages: list[list[Any]] = self._handle_pagination(
                data, table=False, params=params, json=json
            )  # type: ignore[assignment]
            results = [i for page in pages for i in page]
        else:
            results = data
        return [self.model_class(**i) for i in results]

    def _tabulate(
        self,
        params: ParamType | None = {},
        json: JsonType | None = None,
        path: str = "",
    ) -> pd.DataFrame:
        # we can assume this type on table endpoints
        data: dict[str, Any] = self._request_enumeration(
            table=True, params=params, json=json, path=path
        )  # type: ignore[assignment]
        pagination = data.get("pagination", None)
        if pagination is not None:
            # we can assume this type on table endpoints
            pages: list[dict[str, Any]] = self._handle_pagination(
                data, table=True, params=params, json=json, path=path
            )  # type: ignore[assignment]
            dfs = [DataFrame(**page).to_pandas() for page in pages]
            return pd.concat(dfs).replace({np.nan: None})
        else:
            return DataFrame(**data).to_pandas().replace({np.nan: None})

    def _create(
        self, *args: Unpack[tuple[str]], **kwargs: Unpack[_RequestKwargs]
    ) -> dict[str, Any]:
        # we can assume this type on create endpoints
        return self._request("POST", *args, **kwargs)  # type: ignore[return-value]

    def _delete(self, id: int) -> None:
        self._request("DELETE", f"{self.prefix}{str(id)}/")


class GetKwargs(TypedDict, total=False):
    dimension_id: int
    run_ids: list[int]
    parameters: Mapping[str, Any]
    name: str
    run_id: int
    key: str
    model: dict[str, str]
    scenario: dict[str, str]
    version: int
    default_only: bool
    is_default: bool | None


class Retriever(BaseRepository[ModelType]):
    def get(self, **kwargs: Unpack[GetKwargs]) -> ModelType:
        _kwargs = cast(
            dict[
                str,
                bool | int | str | list[int] | Iterable[int] | Mapping[str, Any] | None,
            ],
            kwargs,
        )
        list_ = (
            self._list(params=_kwargs)
            if self.enumeration_method == "GET"
            else self._list(json=_kwargs)
        )

        try:
            [obj] = list_
        except ValueError as e:
            raise self.model_class.NotFound(
                f"Expected exactly one result, got {len(list_)} instead."
            ) from e
        return obj


class Creator(BaseRepository[ModelType]):
    def create(
        self,
        **kwargs: int
        | str
        | Mapping[str, Any]
        | abstract.MetaValue
        | list[str]
        | float
        | None,
    ) -> ModelType:
        res = self._create(self.prefix, json=kwargs)
        return self.model_class(**res)


class Deleter(BaseRepository[ModelType]):
    def delete(self, id: int) -> None:
        self._delete(id)


class ListKwargs(TypedDict, total=False):
    run_id: int
    name: str


class Lister(BaseRepository[ModelType]):
    def list(self, **kwargs: Unpack[ListKwargs]) -> list[ModelType]:
        return self._list(json=kwargs)  # type: ignore[arg-type]


class Tabulator(BaseRepository[ModelType]):
    def tabulate(self, **kwargs: Any) -> pd.DataFrame:
        return self._tabulate(json=kwargs)


class Enumerator(Lister[ModelType], Tabulator[ModelType]):
    def enumerate(
        self, table: bool = False, **kwargs: Any
    ) -> list[ModelType] | pd.DataFrame:
        return self.tabulate(**kwargs) if table else self.list(**kwargs)


class BulkOperator(BaseRepository[ModelType]):
    def yield_chunks(
        self, df: pd.DataFrame, chunk_size: int
    ) -> Generator[pd.DataFrame, Any, None]:
        for _, chunk in df.groupby(df.index // chunk_size):
            yield chunk


class BulkUpsertKwargs(TypedDict, total=False):
    create_related: bool


class BulkUpserter(BulkOperator[ModelType]):
    def bulk_upsert(
        self,
        df: pd.DataFrame,
        chunk_size: int = settings.client_default_upload_chunk_size,
        **kwargs: Unpack[BulkUpsertKwargs],
    ) -> None:
        for chunk in self.yield_chunks(df, chunk_size):
            self.bulk_upsert_chunk(chunk, **kwargs)

    def bulk_upsert_chunk(
        self, df: pd.DataFrame, **kwargs: Unpack[BulkUpsertKwargs]
    ) -> None:
        dict_ = df_to_dict(df)
        json_ = DataFrame(**dict_).model_dump_json()
        self._request(
            "POST",
            self.prefix + "bulk/",
            params=cast(dict[str, bool | None], kwargs),
            content=json_,
        )


class BulkDeleter(BulkOperator[ModelType]):
    def bulk_delete(
        self,
        df: pd.DataFrame,
        chunk_size: int = settings.client_default_upload_chunk_size,
        # NOTE nothing in our code base supplies kwargs here
        **kwargs: Any,
    ) -> None:
        for chunk in self.yield_chunks(df, chunk_size):
            self.bulk_delete_chunk(chunk, **kwargs)

    # NOTE this only gets kwargs from bulk_delete()
    def bulk_delete_chunk(self, df: pd.DataFrame, **kwargs: Any) -> None:
        dict_ = df_to_dict(df)
        json_ = DataFrame(**dict_).model_dump_json()
        self._request(
            "PATCH",
            self.prefix + "bulk/",
            params=kwargs,
            content=json_,
        )



import json
import logging
import logging.config
import sys
from pathlib import Path
from typing import Any, Literal

from httpx import ConnectError
from pydantic import Field, HttpUrl, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from ixmp4 import __file__ as __root__file__
from ixmp4.core.exceptions import InvalidCredentials

from .auth import AnonymousAuth, ManagerAuth
from .credentials import Credentials
from .manager import ManagerConfig
from .toml import TomlConfig
from .user import local_user

logger = logging.getLogger(__name__)

here = Path(__file__).parent
root = Path(__root__file__).parent.parent

try:
    __IPYTHON__  # type: ignore
    _in_ipython_session = True
except NameError:
    _in_ipython_session = False

_sys_has_ps1 = hasattr(sys, "ps1")


class Settings(BaseSettings):
    mode: Literal["production"] | Literal["development"] | Literal["debug"] = (
        "production"
    )
    storage_directory: Path = Field(Path("~/.local/share/ixmp4/"))
    secret_hs256: str = "default_secret_hs256"
    migration_db_uri: str = "sqlite:///./run/db.sqlite"
    manager_url: HttpUrl = Field(HttpUrl("https://api.manager.ece.iiasa.ac.at/v1"))
    managed: bool = True
    max_page_size: int = 10_000
    default_page_size: int = 5_000
    client_default_upload_chunk_size: int = 10_000
    client_max_concurrent_requests: int = Field(2, le=4)
    client_max_request_retries: int = Field(3)
    client_backoff_factor: int = Field(5)
    client_timeout: int = Field(30)
    model_config = SettingsConfigDict(env_prefix="ixmp4_", extra="allow")

    # We don't pass any args or kwargs, so allow all to flow through
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.setup_directories()

        if self.is_in_interactive_mode():
            self.configure_logging(self.mode)

        self._credentials: Credentials | None = None
        self._toml: TomlConfig | None = None
        self._default_auth: ManagerAuth | AnonymousAuth | None = None
        self._manager: ManagerConfig | None = None

        logger.debug(f"Settings loaded: {self}")

    def is_in_interactive_mode(self) -> bool:
        return _sys_has_ps1 or _in_ipython_session

    @property
    def credentials(self) -> Credentials:
        if self._credentials is None:
            self.load_credentials()
        # For this and similar below, mypy doesn't realize that the attribute will not
        # be None after the load() call
        return self._credentials  # type: ignore[return-value]

    @property
    def default_credentials(self) -> tuple[str, str] | None:
        try:
            return self.credentials.get("default")
        except KeyError:
            return None

    @property
    def toml(self) -> TomlConfig:
        if self._toml is None:
            self.load_toml_config()
        return self._toml  # type: ignore[return-value]

    @property
    def default_auth(self) -> ManagerAuth | AnonymousAuth | None:
        if self._default_auth is None:
            self.get_auth()
        return self._default_auth

    @property
    def manager(self) -> ManagerConfig:
        if self._manager is None:
            self.load_manager_config()
        return self._manager  # type: ignore[return-value]

    def setup_directories(self) -> None:
        self.storage_directory = Path.expanduser(self.storage_directory)

        if not self.storage_directory.is_absolute():
            self.storage_directory = root / self.storage_directory

        self.storage_directory.mkdir(parents=True, exist_ok=True)

        self.database_dir = self.storage_directory / "databases"
        self.database_dir.mkdir(exist_ok=True)

        self.log_dir = self.storage_directory / "log"
        self.log_dir.mkdir(exist_ok=True)

    def load_credentials(self) -> None:
        credentials_config = self.storage_directory / "credentials.toml"
        credentials_config.touch()
        self._credentials = Credentials(credentials_config)

    def get_auth(self) -> None:
        if self.default_credentials is not None:
            try:
                self._default_auth = ManagerAuth(
                    *self.default_credentials, str(self.manager_url)
                )
                logger.info(
                    f"Connecting as user '{self._default_auth.get_user().username}'."
                )
            except InvalidCredentials:
                logger.warning(f"Invalid credentials for {self.manager_url}.")
            except ConnectError:
                logger.warning(f"Unable to connect to {self.manager_url}.")

        else:
            self._default_auth = AnonymousAuth()

    def load_manager_config(self) -> None:
        self._manager = ManagerConfig(
            str(self.manager_url), self.default_auth, remote=True
        )

    def load_toml_config(self) -> None:
        if self.default_auth is not None:
            toml_user = self.default_auth.get_user()
            if not toml_user.is_authenticated:
                toml_user = local_user
        else:  # if no connection to manager
            toml_user = local_user

        toml_config = self.storage_directory / "platforms.toml"
        toml_config.touch()
        self._toml = TomlConfig(toml_config, toml_user)

    @field_validator("storage_directory")
    def expand_user(cls, v: Path) -> Path:
        # translate ~/asdf into /home/user/asdf
        return Path.expanduser(v)

    def get_server_logconf(self) -> Path:
        return here / "./logging/server.json"

    def configure_logging(self, config: str) -> None:
        self.access_file = str((self.log_dir / "access.log").absolute())
        self.debug_file = str((self.log_dir / "debug.log").absolute())
        self.error_file = str((self.log_dir / "error.log").absolute())

        logging_config = here / f"logging/{config}.json"
        with open(logging_config) as file:
            config_dict = json.load(file)
        logging.config.dictConfig(config_dict)

    def check_credentials(self) -> None:
        if self.default_credentials is not None:
            username, password = self.default_credentials
            ManagerAuth(username, password, str(self.manager_url))



from collections.abc import Iterable
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, cast

import pandas as pd

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from .. import base
from ..docs import Docs, DocsRepository

if TYPE_CHECKING:
    from ixmp4.data.backend.api import RestBackend


class Table(base.BaseModel):
    NotFound: ClassVar = abstract.Table.NotFound
    NotUnique: ClassVar = abstract.Table.NotUnique
    DeletionPrevented: ClassVar = abstract.Table.DeletionPrevented

    id: int
    name: str
    data: dict[str, list[float] | list[int] | list[str]]
    indexset_names: list[str]
    column_names: list[str] | None
    run__id: int

    created_at: datetime | None
    created_by: str | None


class TableDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/optimization/tables/"


class TableRepository(
    base.Creator[Table],
    base.Deleter[Table],
    base.Retriever[Table],
    base.Enumerator[Table],
    abstract.TableRepository,
):
    model_class = Table
    prefix = "optimization/tables/"

    def __init__(self, *args: Unpack[tuple["RestBackend"]]) -> None:
        super().__init__(*args)
        self.docs = TableDocsRepository(self.backend)

    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Table:
        return super().create(
            name=name,
            run_id=run_id,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )

    def delete(self, id: int) -> None:
        super().delete(id=id)

    def add_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, pd.DataFrame):
            # data will always contain str, not only Hashable
            dict_data = cast(dict[str, Any], data.to_dict(orient="list"))
            data = dict_data
        kwargs = {"data": data}
        self._request(
            method="PATCH", path=self.prefix + str(id) + "/data/", json=kwargs
        )

    def remove_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, pd.DataFrame):
            # data will always contain str, not only Hashable
            dict_data = cast(dict[str, Any], data.to_dict(orient="list"))
            data = dict_data
        kwargs = {"data": data}
        self._request(
            method="DELETE", path=self.prefix + str(id) + "/data/", json=kwargs
        )

    def get(self, run_id: int, name: str) -> Table:
        return super().get(run_id=run_id, name=name)

    def get_by_id(self, id: int) -> Table:
        res = self._get_by_id(id)
        return Table(**res)

    def list(
        self, **kwargs: Unpack[abstract.optimization.EnumerateKwargs]
    ) -> Iterable[Table]:
        json = cast(abstract.annotations.OptimizationFilterAlias, kwargs)
        return super()._list(json=json)

    def tabulate(
        self, **kwargs: Unpack[abstract.optimization.EnumerateKwargs]
    ) -> pd.DataFrame:
        json = cast(abstract.annotations.OptimizationFilterAlias, kwargs)
        return super()._tabulate(json=json)

    def enumerate(
        self, **kwargs: Unpack[abstract.optimization.EnumerateKwargs]
    ) -> Iterable[Table] | pd.DataFrame:
        return super().enumerate(**kwargs)



from collections.abc import Iterable
from typing import Any

from ixmp4.data.auth.decorators import guard

from .. import base
from ..docs import AbstractDocs, BaseDocsRepository, docs_model
from .model import Unit

UnitDocs = docs_model(Unit)


class UnitDocsRepository(
    BaseDocsRepository[Any],
    base.BaseRepository[Unit],
):
    model_class = UnitDocs
    dimension_model_class = Unit

    @guard("view")
    def list(
        self,
        *,
        dimension_id: int | None = None,
        dimension_id__in: Iterable[int] | None = None,
    ) -> list[AbstractDocs]:
        return super().list(
            dimension_id=dimension_id, dimension_id__in=dimension_id__in
        )



from .datapoint import (  # AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint
    DataPoint,
    DataPointRepository,
    UniversalDataPoint,
)
from .measurand import Measurand, MeasurandRepository
from .timeseries import TimeSeries, TimeSeriesRepository
from .variable import Variable, VariableRepository



import logging
from collections.abc import Callable, Generator, Mapping, Sequence
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, cast

from sqlalchemy import Connection, Result, event, sql
from sqlalchemy.orm import Mapper, ORMExecuteState

from ixmp4 import db
from ixmp4.core.exceptions import ProgrammingError
from ixmp4.data.db import base, mixins

if TYPE_CHECKING:
    from ..backend.db import SqlAlchemyBackend


ParametersT = Sequence[Mapping[str, Any]] | Mapping[str, Any]


class SqlaEventHandler(object):
    backend: "SqlAlchemyBackend"

    def __init__(self, backend: "SqlAlchemyBackend") -> None:
        self.backend = backend
        self.listeners: list[
            tuple[tuple[Any, str, Callable[..., None]], dict[str, bool]]
        ] = [
            ((backend.session, "do_orm_execute", self.receive_do_orm_execute), {}),
            (
                (base.BaseModel, "before_insert", self.receive_before_insert),
                {"propagate": True},
            ),
            (
                (base.BaseModel, "before_update", self.receive_before_update),
                {"propagate": True},
            ),
        ]
        self.add_listeners()

    def add_listeners(self) -> None:
        # Somehow, mypy knows how long args is, but tries to insert kwargs as argument
        # 2 and 3
        for args, kwargs in self.listeners:
            if not event.contains(*args):
                event.listen(*args, **kwargs)  # type: ignore[arg-type]

    def remove_listeners(self) -> None:
        for args, kwargs in self.listeners:
            if event.contains(*args):
                event.remove(*args)

    @contextmanager
    def pause(self) -> Generator[None, Any, None]:
        """Temporarily removes all event listeners for the enclosed scope."""
        self.remove_listeners()
        yield
        self.add_listeners()

    def set_logger(self, state: tuple[int, int, int] | ORMExecuteState) -> None:
        self.logger = logging.getLogger(__name__ + "." + str(id(state)))

    def receive_before_insert(
        self, mapper: Mapper[Any], connection: Connection, target: base.BaseModel
    ) -> None:
        """Handles the insert event when creating data like this:
        ```
        model = Model(**kwargs)
        session.add(model)
        session.commit()
        ```
        """
        self.set_logger((id(mapper), id(connection), id(target)))
        if connection.engine is not self.backend.engine:
            self.logger.debug("Event dispatched from another backend, ignoring.")
            return

        if isinstance(target, mixins.HasCreationInfo):
            self.logger.debug(f"Setting creation info for: {target}")
            target.set_creation_info(self.backend.auth_context)

    def receive_before_update(
        self, mapper: Mapper[Any], connection: Connection, target: base.BaseModel
    ) -> None:
        """Handles the update event when changing data like this:
        ```
        model = query_model()
        model.attribute = "foo"
        session.commit()
        ```
        """
        self.set_logger((id(mapper), id(connection), id(target)))
        if connection.engine is not self.backend.engine:
            self.logger.debug("Event dispatched from another backend, ignoring.")
            return

        if isinstance(target, mixins.HasUpdateInfo):
            self.logger.debug(f"Setting update info for: {target}")
            target.set_update_info(self.backend.auth_context)

    def receive_do_orm_execute(self, orm_execute_state: ORMExecuteState) -> None:
        """Handles ORM execution events like:
        ```
        exc = select/update/delete(Model)
        exc = exc.where(<expression>)
        result = session.execute(exc)
        # or
        exc = insert/update(Model)
        result = session.execute(exc, [{"attribute": "foo", ...}, ...])
        ```
        """
        self.set_logger(orm_execute_state)
        self.logger.debug("Received 'do_orm_execute' event.")
        if orm_execute_state.is_select:
            self.receive_select(orm_execute_state)
        else:
            if orm_execute_state.is_insert:
                self.logger.debug("Operation: 'insert'")
                self.receive_insert(orm_execute_state)
            if orm_execute_state.is_update:
                self.logger.debug("Operation: 'update'")
                self.receive_update(orm_execute_state)
            if orm_execute_state.is_delete:
                self.logger.debug("Operation: 'delete'")
                self.receive_delete(orm_execute_state)
            else:
                self.logger.debug(f"Ignoring operation: {orm_execute_state}")

    def receive_select(self, oes: ORMExecuteState) -> None:
        # select = cast(sql.Select, oes.statement)
        pass

    def receive_insert(self, oes: ORMExecuteState) -> Result[Any] | None:
        insert = cast(sql.Insert, oes.statement)
        entity = insert.entity_description
        type_ = entity["type"]
        self.logger.debug(f"Entity: '{entity['name']}'")

        if issubclass(type_, mixins.HasCreationInfo):
            creation_info = {
                "created_by": type_.get_username(self.backend.auth_context),
                "created_at": type_.get_timestamp(),
            }

            assert oes.parameters is not None
            return oes.invoke_statement(
                params=self.get_extra_params(oes.parameters, creation_info)
            )
        return None

    def receive_update(self, oes: ORMExecuteState) -> Result[Any] | None:
        update = cast(sql.Update, oes.statement)
        entity = update.entity_description
        type_ = entity["type"]
        self.logger.debug(f"Entity: '{entity['name']}'")

        if issubclass(type_, mixins.HasUpdateInfo):
            update_info = {
                "updated_by": type_.get_username(self.backend.auth_context),
                "updated_at": type_.get_timestamp(),
            }
            if oes.parameters is not None:
                return oes.invoke_statement(
                    params=self.get_extra_params(oes.parameters, update_info)
                )
            elif update.whereclause is not None:
                new_statement = update.values(**update_info)
                return oes.invoke_statement(statement=new_statement)
            else:
                raise ProgrammingError(f"Cannot handle update statement: {update}")
        return None

    def receive_delete(self, oes: ORMExecuteState) -> None:
        # delete = cast(sql.Delete, oes.statement)
        pass

    # Can't find any reference use cases
    def select_affected(self, statement: sql.Delete | sql.Update) -> sql.Select[Any]:
        entity = statement.entity_description
        wc = statement.whereclause
        exc = db.select(entity["entity"])
        if wc is not None:
            exc.where(wc.expression)
        return exc

    def get_extra_params(
        self,
        params: Sequence[Mapping[str, Any]] | Mapping[str, Any],
        extra: dict[str, Any],
    ) -> list[dict[str, Any]] | dict[str, Any]:
        return [extra] * len(params) if isinstance(params, Sequence) else extra



from fastapi import APIRouter, Body, Depends, Path, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.checkpoint import Checkpoint, CheckpointFilter

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination
from .decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/checkpoints",
    tags=["checkpoints"],
)


class CheckpointInput(BaseModel):
    run__id: int
    message: str


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.Checkpoint])
def query(
    filter: CheckpointFilter = Body(None),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
) -> EnumerationOutput[Checkpoint]:
    return EnumerationOutput(
        results=backend.checkpoints.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.units.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.post("/", response_model=api.Checkpoint)
def create(
    checkpoint: CheckpointInput = Body(),
    backend: Backend = Depends(deps.get_backend),
) -> Checkpoint:
    return backend.checkpoints.create(checkpoint.run__id, checkpoint.message)


@autodoc
@router.delete("/{id}/")
def delete(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.checkpoints.delete(id)



from datetime import datetime
from typing import Literal

from sqlalchemy.orm import Mapped as Mapped

from ixmp4 import db

Boolean = Mapped[bool]
DateTime = Mapped[datetime]
Float = Mapped[float]
EquationId = Mapped[db.EquationIdType]
IndexSetId = Mapped[db.IndexSetIdType]
IndexSet__Id = Mapped[db.IndexSet__IdType]
Id = Mapped[int]
Integer = Mapped[int]
JsonDict = Mapped[dict[str, list[float] | list[int] | list[str]]]
Name = Mapped[db.NameType]
OptimizationDataList = Mapped[list[float] | list[int] | list[str]]
OptimizationDataType = Mapped[Literal["float", "int", "str"] | None]
OptimizationVariableType = Mapped[db.OptimizationVariableIdType]
ParameterId = Mapped[db.ParameterIdType]
RunId = Mapped[db.RunIdType]
String = Mapped[str]
TableId = Mapped[db.TableIdType]
UniqueName = Mapped[db.UniqueNameType]
Username = Mapped[db.UsernameType]



from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Protocol, _ProtocolMeta

from typing_extensions import Unpack

if TYPE_CHECKING:
    from ixmp4.data.db.base import EnumerateKwargs

import pandas as pd

from ixmp4.core.exceptions import DeletionPrevented, IxmpError, NotFound, NotUnique


class BaseMeta(_ProtocolMeta):
    def __init__(
        self, name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> None:
        super().__init__(name, bases, namespace)
        self.NotUnique = type(
            self.__name__ + "NotUnique",
            (NotUnique,),
            {"http_error_name": name.lower() + "_not_unique"},
        )
        self.NotFound = type(
            self.__name__ + "NotFound",
            (NotFound,),
            {"http_error_name": name.lower() + "_not_found"},
        )

        self.DeletionPrevented = type(
            self.__name__ + "DeletionPrevented",
            (DeletionPrevented,),
            {"http_error_name": name.lower() + "_deletion_prevented"},
        )


class BaseModel(Protocol, metaclass=BaseMeta):
    NotUnique: ClassVar[type[IxmpError]]
    NotFound: ClassVar[type[IxmpError]]
    DeletionPrevented: ClassVar[type[IxmpError]]

    id: int
    "Unique integer id."


class BaseRepository(Protocol):
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...


class Retriever(BaseRepository, Protocol):
    def get(self, *args: Any, **kwargs: Any) -> BaseModel: ...


class Creator(BaseRepository, Protocol):
    def create(self, *args: Any, **kwargs: Any) -> BaseModel: ...


class Deleter(BaseRepository, Protocol):
    def delete(self, *args: Any, **kwargs: Any) -> None: ...


class Lister(BaseRepository, Protocol):
    def list(self, *args: Any, **kwargs: Any) -> Sequence[BaseModel]: ...


class Tabulator(BaseRepository, Protocol):
    def tabulate(self, *args: Any, **kwargs: Any) -> pd.DataFrame: ...


class Enumerator(Lister, Tabulator, Protocol):
    def enumerate(
        self, table: bool = False, **kwargs: Unpack["EnumerateKwargs"]
    ) -> Sequence[BaseModel] | pd.DataFrame: ...


class BulkUpserter(BaseRepository, Protocol):
    def bulk_upsert(self, *args: Any, **kwargs: Any) -> None: ...


class BulkDeleter(BaseRepository, Protocol):
    def bulk_delete(self, *args: Any, **kwargs: Any) -> None: ...



from ixmp4.data.db import filters as base
from ixmp4.data.db.base import SelectType
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.db import Session, filters, utils

from ..measurand import Measurand
from . import Variable


class VariableFilter(base.VariableFilter, metaclass=filters.FilterMeta):
    region: base.RegionFilter | None = filters.Field(None)
    unit: base.UnitFilter | None = filters.Field(None)
    run: base.RunFilter = filters.Field(
        default=base.RunFilter(id=None, version=None, is_default=True)
    )

    _remote_filters = {"run", "region", "unit"}
    _remote_path = [
        {
            "target_model": Measurand,
            "fk_attr": "variable__id",
            "source_model": Variable,
            "pk_attr": "id",
        },
        {
            "target_model": TimeSeries,
            "fk_attr": "measurand__id",
            "source_model": Measurand,
            "pk_attr": "id",
        },
    ]

    def join(self, exc: SelectType, session: Session | None = None) -> SelectType:
        if self._should_use_subquery_optimization():
            return exc

        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, Measurand.variable__id == Variable.id)

        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(
                TimeSeries, onclause=TimeSeries.measurand__id == Measurand.id
            )
        return exc



from typing import TYPE_CHECKING

import pandas as pd
from fastapi.testclient import TestClient
from sqlalchemy.engine import create_engine
from sqlalchemy.pool import NullPool, StaticPool

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import Unpack

from ixmp4.conf import settings
from ixmp4.conf.auth import BaseAuth, SelfSignedAuth
from ixmp4.conf.base import PlatformInfo
from ixmp4.conf.manager import ManagerPlatformInfo, MockManagerConfig
from ixmp4.conf.user import User
from ixmp4.server import app, v1
from ixmp4.server.rest import deps

from .api import RestBackend
from .db import SqlAlchemyBackend

if TYPE_CHECKING:
    from .db import SqlAlchemyBackend


class SqliteTestBackend(SqlAlchemyBackend):
    def __init__(self, *args: Unpack[tuple[PlatformInfo]]) -> None:
        super().__init__(*args)

    def make_engine(self, dsn: str) -> None:
        self.engine = create_engine(
            dsn,
            connect_args={"check_same_thread": False},
            poolclass=StaticPool,
            max_identifier_length=63,
        )
        self.session = self.Session(bind=self.engine)


class PostgresTestBackend(SqlAlchemyBackend):
    def __init__(self, *args: Unpack[tuple[PlatformInfo]]) -> None:
        super().__init__(*args)

    def make_engine(self, dsn: str) -> None:
        self.engine = create_engine(
            dsn,
            poolclass=NullPool,
            max_identifier_length=63,
        )
        self.session = self.Session(bind=self.engine)


test_platform = ManagerPlatformInfo(
    id=1,
    management_group=1,
    access_group=1,
    accessibility=ManagerPlatformInfo.Accessibilty.PRIVATE,
    slug="test",
    notice="Welcome to the test platform.",
    dsn="http://testserver/v1/:test:/",
    url="http://testserver/v1/:test:/",
)

test_user = User(id=-1, username="@unknown", is_superuser=True, is_verified=True)

test_permissions = pd.DataFrame(
    [], columns=["id", "instance", "group", "access_type", "model"]
)

mock_manager = MockManagerConfig([test_platform], test_permissions)


class RestTestBackend(RestBackend):
    db_backend: "SqlAlchemyBackend"

    def __init__(self, db_backend: "SqlAlchemyBackend") -> None:
        self.db_backend = db_backend
        self.auth_params = (test_user, mock_manager, test_platform)
        super().__init__(test_platform, SelfSignedAuth(settings.secret_hs256))

    def make_client(self, rest_url: str, auth: BaseAuth) -> None:
        self.client = TestClient(
            app=app,
            base_url=rest_url,
            raise_server_exceptions=False,
        )

        app.dependency_overrides[deps.validate_token] = deps.do_not_validate_token
        v1.dependency_overrides[deps.validate_token] = deps.do_not_validate_token

        app.dependency_overrides[deps.get_backend] = deps.get_test_backend_dependency(
            self.db_backend, self.auth_params
        )
        v1.dependency_overrides[deps.get_backend] = deps.get_test_backend_dependency(
            self.db_backend, self.auth_params
        )

    def close(self) -> None:
        self.client.close()
        self.executor.shutdown(cancel_futures=True)

    def setup(self) -> None:
        pass

    def teardown(self) -> None:
        pass



# type: ignore
"""Enable versioning for optimization items

Revision ID: e189ffe6a65e
Revises: fbda0eb90f37
Create Date: 2025-09-05 19:06:29.177463

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# Revision identifiers, used by Alembic.
revision = "e189ffe6a65e"
down_revision = "fbda0eb90f37"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "opt_idx",
        sa.Column(
            "_data_type",
            sa.Enum("float", "int", "str", native_enum=False),
            nullable=True,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_opt_idx_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_idx")),
        sa.UniqueConstraint("name", "run__id", name=op.f("uq_opt_idx_name_run__id")),
    )
    with op.batch_alter_table("opt_idx", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_idx_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "opt_equ",
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_opt_equ_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_equ")),
        sa.UniqueConstraint("name", "run__id", name=op.f("uq_opt_equ_name_run__id")),
    )
    with op.batch_alter_table("opt_equ", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_equ_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "opt_par",
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_opt_par_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_par")),
        sa.UniqueConstraint("name", "run__id", name=op.f("uq_opt_par_name_run__id")),
    )
    with op.batch_alter_table("opt_par", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_par_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "opt_sca",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Float(), nullable=True),
        sa.Column("unit__id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_opt_sca_run__id_run")
        ),
        sa.ForeignKeyConstraint(
            ["unit__id"], ["unit.id"], name=op.f("fk_opt_sca_unit__id_unit")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_sca")),
        sa.UniqueConstraint("name", "run__id", name=op.f("uq_opt_sca_name_run__id")),
    )
    with op.batch_alter_table("opt_sca", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_sca_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_opt_sca_unit__id"), ["unit__id"], unique=False
        )

    op.create_table(
        "opt_tab",
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_opt_tab_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_tab")),
        sa.UniqueConstraint("name", "run__id", name=op.f("uq_opt_tab_name_run__id")),
    )
    with op.batch_alter_table("opt_tab", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_tab_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "opt_var",
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_opt_var_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_var")),
        sa.UniqueConstraint("name", "run__id", name=op.f("uq_opt_var_name_run__id")),
    )
    with op.batch_alter_table("opt_var", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_var_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "opt_equ_idx_association_version",
        sa.Column("equation__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_equ_idx_association_version")
        ),
    )
    with op.batch_alter_table(
        "opt_equ_idx_association_version", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_equ_idx_association_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_idx_association_version_equation__id"),
            ["equation__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_idx_association_version_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_idx_association_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_idx_association_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_equ_version",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_equ_version")
        ),
    )
    with op.batch_alter_table("opt_equ_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_equ_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_idx_data_version",
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("value", sa.String(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_idx_data_version")
        ),
    )
    with op.batch_alter_table("opt_idx_data_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_idx_data_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_idx_data_version_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_idx_data_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_idx_data_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_idx_version",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "_data_type",
            sa.Enum("float", "int", "str", native_enum=False),
            nullable=True,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_idx_version")
        ),
    )
    with op.batch_alter_table("opt_idx_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_idx_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_idx_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_idx_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_opt_idx_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_par_idx_association_version",
        sa.Column("parameter__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_par_idx_association_version")
        ),
    )
    with op.batch_alter_table(
        "opt_par_idx_association_version", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_par_idx_association_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_idx_association_version_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_idx_association_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_idx_association_version_parameter__id"),
            ["parameter__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_idx_association_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_par_version",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_par_version")
        ),
    )
    with op.batch_alter_table("opt_par_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_par_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_sca_version",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Float(), nullable=True),
        sa.Column("unit__id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_sca_version")
        ),
    )
    with op.batch_alter_table("opt_sca_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_sca_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_sca_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_sca_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_opt_sca_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_sca_version_unit__id"), ["unit__id"], unique=False
        )

    op.create_table(
        "opt_tab_idx_association_version",
        sa.Column("table__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_tab_idx_association_version")
        ),
    )
    with op.batch_alter_table(
        "opt_tab_idx_association_version", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_tab_idx_association_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_idx_association_version_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_idx_association_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_idx_association_version_table__id"),
            ["table__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_idx_association_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_tab_version",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_tab_version")
        ),
    )
    with op.batch_alter_table("opt_tab_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_tab_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_var_idx_association_version",
        sa.Column("variable__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_var_idx_association_version")
        ),
    )
    with op.batch_alter_table(
        "opt_var_idx_association_version", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_var_idx_association_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_idx_association_version_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_idx_association_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_idx_association_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_idx_association_version_variable__id"),
            ["variable__id"],
            unique=False,
        )

    op.create_table(
        "opt_var_version",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.BigInteger(), nullable=False),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_opt_var_version")
        ),
    )
    with op.batch_alter_table("opt_var_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_var_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "opt_equ_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["opt_equ.id"],
            name=op.f("fk_opt_equ_docs_dimension__id_opt_equ"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_equ_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_opt_equ_docs_dimension__id")
        ),
    )
    op.create_table(
        "opt_equ_idx_association",
        sa.Column("equation__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["equation__id"],
            ["opt_equ.id"],
            name=op.f("fk_opt_equ_idx_association_equation__id_opt_equ"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["opt_idx.id"],
            name=op.f("fk_opt_equ_idx_association_indexset__id_opt_idx"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_equ_idx_association")),
    )
    with op.batch_alter_table("opt_equ_idx_association", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_equ_idx_association_equation__id"),
            ["equation__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_equ_idx_association_indexset__id"),
            ["indexset__id"],
            unique=False,
        )

    op.create_table(
        "opt_idx_data",
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("value", sa.String(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["opt_idx.id"],
            name=op.f("fk_opt_idx_data_indexset__id_opt_idx"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_idx_data")),
        sa.UniqueConstraint(
            "indexset__id", "value", name=op.f("uq_opt_idx_data_indexset__id_value")
        ),
    )
    with op.batch_alter_table("opt_idx_data", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_idx_data_indexset__id"), ["indexset__id"], unique=False
        )

    op.create_table(
        "opt_idx_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["opt_idx.id"],
            name=op.f("fk_opt_idx_docs_dimension__id_opt_idx"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_idx_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_opt_idx_docs_dimension__id")
        ),
    )
    op.create_table(
        "opt_par_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["opt_par.id"],
            name=op.f("fk_opt_par_docs_dimension__id_opt_par"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_par_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_opt_par_docs_dimension__id")
        ),
    )
    op.create_table(
        "opt_par_idx_association",
        sa.Column("parameter__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["opt_idx.id"],
            name=op.f("fk_opt_par_idx_association_indexset__id_opt_idx"),
        ),
        sa.ForeignKeyConstraint(
            ["parameter__id"],
            ["opt_par.id"],
            name=op.f("fk_opt_par_idx_association_parameter__id_opt_par"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_par_idx_association")),
    )
    with op.batch_alter_table("opt_par_idx_association", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_par_idx_association_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_par_idx_association_parameter__id"),
            ["parameter__id"],
            unique=False,
        )

    op.create_table(
        "opt_sca_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["opt_sca.id"],
            name=op.f("fk_opt_sca_docs_dimension__id_opt_sca"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_sca_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_opt_sca_docs_dimension__id")
        ),
    )
    op.create_table(
        "opt_tab_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["opt_tab.id"],
            name=op.f("fk_opt_tab_docs_dimension__id_opt_tab"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_tab_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_opt_tab_docs_dimension__id")
        ),
    )
    op.create_table(
        "opt_tab_idx_association",
        sa.Column("table__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["opt_idx.id"],
            name=op.f("fk_opt_tab_idx_association_indexset__id_opt_idx"),
        ),
        sa.ForeignKeyConstraint(
            ["table__id"],
            ["opt_tab.id"],
            name=op.f("fk_opt_tab_idx_association_table__id_opt_tab"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_tab_idx_association")),
    )
    with op.batch_alter_table("opt_tab_idx_association", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_tab_idx_association_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_tab_idx_association_table__id"),
            ["table__id"],
            unique=False,
        )

    op.create_table(
        "opt_var_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["opt_var.id"],
            name=op.f("fk_opt_var_docs_dimension__id_opt_var"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_var_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_opt_var_docs_dimension__id")
        ),
    )
    op.create_table(
        "opt_var_idx_association",
        sa.Column("variable__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["opt_idx.id"],
            name=op.f("fk_opt_var_idx_association_indexset__id_opt_idx"),
        ),
        sa.ForeignKeyConstraint(
            ["variable__id"],
            ["opt_var.id"],
            name=op.f("fk_opt_var_idx_association_variable__id_opt_var"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_opt_var_idx_association")),
    )
    with op.batch_alter_table("opt_var_idx_association", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_opt_var_idx_association_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_opt_var_idx_association_variable__id"),
            ["variable__id"],
            unique=False,
        )

    with op.batch_alter_table(
        "optimization_tableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_tableindexsetassociation_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_tableindexsetassociation_table__id")
        )

    op.drop_table("optimization_tableindexsetassociation")
    op.drop_table("optimization_parameter_docs")
    op.drop_table("optimization_equation_docs")
    with op.batch_alter_table(
        "optimization_variableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_variableindexsetassociation_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_variableindexsetassociation_variable__id")
        )

    op.drop_table("optimization_variableindexsetassociation")
    op.drop_table("optimization_table_docs")
    op.drop_table("optimization_optimizationvariable_docs")
    with op.batch_alter_table("optimization_indexsetdata", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_indexsetdata_indexset__id"))

    op.drop_table("optimization_indexsetdata")
    op.drop_table("optimization_indexset_docs")
    with op.batch_alter_table(
        "optimization_equationindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_equationindexsetassociation_equation__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_equationindexsetassociation_indexset__id")
        )

    op.drop_table("optimization_equationindexsetassociation")
    with op.batch_alter_table(
        "optimization_parameterindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_parameter__id")
        )

    op.drop_table("optimization_parameterindexsetassociation")
    op.drop_table("optimization_scalar_docs")

    with op.batch_alter_table("optimization_scalar", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_scalar_run__id"))
        batch_op.drop_index(batch_op.f("ix_optimization_scalar_unit__id"))

    op.drop_table("optimization_scalar")

    with op.batch_alter_table("optimization_table", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_table_run__id"))

    op.drop_table("optimization_table")

    with op.batch_alter_table("optimization_parameter", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_parameter_run__id"))

    op.drop_table("optimization_parameter")

    with op.batch_alter_table("optimization_equation", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_equation_run__id"))

    op.drop_table("optimization_equation")

    with op.batch_alter_table(
        "optimization_optimizationvariable", schema=None
    ) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_optimizationvariable_run__id"))

    op.drop_table("optimization_optimizationvariable")

    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_indexset_run__id"))

    op.drop_table("optimization_indexset")

    with op.batch_alter_table("iamc_variable_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_iamc_variable_docs_dimension__id_iamc_variable"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_iamc_variable_docs_dimension__id_iamc_variable"),
            "iamc_variable",
            ["dimension__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("model_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_model_docs_dimension__id_model"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_model_docs_dimension__id_model"),
            "model",
            ["dimension__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("region_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_region_docs_dimension__id_region"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_region_docs_dimension__id_region"),
            "region",
            ["dimension__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("scenario_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_scenario_docs_dimension__id_scenario"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_scenario_docs_dimension__id_scenario"),
            "scenario",
            ["dimension__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("unit_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_unit_docs_dimension__id_unit"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_unit_docs_dimension__id_unit"),
            "unit",
            ["dimension__id"],
            ["id"],
            ondelete="CASCADE",
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("unit_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_unit_docs_dimension__id_unit"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_unit_docs_dimension__id_unit"),
            "unit",
            ["dimension__id"],
            ["id"],
        )

    with op.batch_alter_table("scenario_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_scenario_docs_dimension__id_scenario"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_scenario_docs_dimension__id_scenario"),
            "scenario",
            ["dimension__id"],
            ["id"],
        )

    with op.batch_alter_table("region_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_region_docs_dimension__id_region"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_region_docs_dimension__id_region"),
            "region",
            ["dimension__id"],
            ["id"],
        )

    with op.batch_alter_table("model_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_model_docs_dimension__id_model"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_model_docs_dimension__id_model"),
            "model",
            ["dimension__id"],
            ["id"],
        )

    with op.batch_alter_table("iamc_variable_docs", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_iamc_variable_docs_dimension__id_iamc_variable"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_iamc_variable_docs_dimension__id_iamc_variable"),
            "iamc_variable",
            ["dimension__id"],
            ["id"],
        )

    op.create_table(
        "optimization_indexset",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("_data_type", sa.VARCHAR(length=5), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name="fk_optimization_indexset_run__id_run"
        ),
        sa.PrimaryKeyConstraint("id", name="pk_optimization_indexset"),
        sa.UniqueConstraint(
            "name", "run__id", name="uq_optimization_indexset_name_run__id"
        ),
    )
    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_indexset_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_equation",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_optimization_equation_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_equation")),
        sa.UniqueConstraint(
            "name", "run__id", name=op.f("uq_optimization_equation_name_run__id")
        ),
    )
    with op.batch_alter_table("optimization_equation", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_equation_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_table",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name="fk_optimization_table_run__id_run"
        ),
        sa.PrimaryKeyConstraint("id", name="pk_optimization_table"),
        sa.UniqueConstraint(
            "name", "run__id", name="uq_optimization_table_name_run__id"
        ),
    )
    with op.batch_alter_table("optimization_table", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_table_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_scalar",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("value", sa.Float(), nullable=True),
        sa.Column("unit__id", sa.Integer(), nullable=False),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_optimization_scalar_run__id_run")
        ),
        sa.ForeignKeyConstraint(
            ["unit__id"], ["unit.id"], name=op.f("fk_optimization_scalar_unit__id_unit")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_scalar")),
        sa.UniqueConstraint(
            "name", "run__id", name=op.f("uq_optimization_scalar_name_run__id")
        ),
    )
    with op.batch_alter_table("optimization_scalar", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_scalar_unit__id"), ["unit__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_scalar_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_optimizationvariable",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"],
            ["run.id"],
            name="fk_optimization_optimizationvariable_run__id_run",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_optimization_optimizationvariable"),
        sa.UniqueConstraint(
            "name", "run__id", name="uq_optimization_optimizationvariable_name_run__id"
        ),
    )
    with op.batch_alter_table(
        "optimization_optimizationvariable", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_optimizationvariable_run__id"),
            ["run__id"],
            unique=False,
        )

    op.create_table(
        "optimization_parameter",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name="fk_optimization_parameter_run__id_run"
        ),
        sa.PrimaryKeyConstraint("id", name="pk_optimization_parameter"),
        sa.UniqueConstraint(
            "name", "run__id", name="uq_optimization_parameter_name_run__id"
        ),
    )
    with op.batch_alter_table("optimization_parameter", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_parameter_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_scalar_docs",
        sa.Column("description", sa.TEXT(), nullable=False),
        sa.Column("dimension__id", sa.INTEGER(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_scalar.id"],
            name=op.f("fk_optimization_scalar_docs_dimension__id_optimization_scalar"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_scalar_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_scalar_docs_dimension__id")
        ),
    )
    op.create_table(
        "optimization_parameterindexsetassociation",
        sa.Column("parameter__id", sa.INTEGER(), nullable=False),
        sa.Column("indexset__id", sa.INTEGER(), nullable=False),
        sa.Column("column_name", sa.VARCHAR(length=255), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_parameterindexsetassociation_indexset___7045"),
        ),
        sa.ForeignKeyConstraint(
            ["parameter__id"],
            ["optimization_parameter.id"],
            name=op.f("fk_optimization_parameterindexsetassociation_parameter__8a6d"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_parameterindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_parameterindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_parameter__id"),
            ["parameter__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )

    op.create_table(
        "optimization_equationindexsetassociation",
        sa.Column("equation__id", sa.INTEGER(), nullable=False),
        sa.Column("indexset__id", sa.INTEGER(), nullable=False),
        sa.Column("column_name", sa.VARCHAR(length=255), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["equation__id"],
            ["optimization_equation.id"],
            name=op.f("fk_optimization_equationindexsetassociation_equation__i_be8b"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_equationindexsetassociation_indexset__i_82d0"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_equationindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_equationindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_equationindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_equationindexsetassociation_equation__id"),
            ["equation__id"],
            unique=False,
        )

    op.create_table(
        "optimization_indexset_docs",
        sa.Column("description", sa.TEXT(), nullable=False),
        sa.Column("dimension__id", sa.INTEGER(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_indexset_docs_dimension__id_optimizatio_72ee"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_indexset_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_indexset_docs_dimension__id")
        ),
    )

    op.create_table(
        "optimization_indexsetdata",
        sa.Column("indexset__id", sa.INTEGER(), nullable=False),
        sa.Column("value", sa.VARCHAR(), nullable=False),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f(
                "fk_optimization_indexsetdata_indexset__id_optimization_indexset"
            ),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_indexsetdata")),
        sa.UniqueConstraint(
            "indexset__id",
            "value",
            name=op.f("uq_optimization_indexsetdata_indexset__id_value"),
        ),
    )
    with op.batch_alter_table("optimization_indexsetdata", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_indexsetdata_indexset__id"),
            ["indexset__id"],
            unique=False,
        )

    op.create_table(
        "optimization_optimizationvariable_docs",
        sa.Column("description", sa.TEXT(), nullable=False),
        sa.Column("dimension__id", sa.INTEGER(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_optimizationvariable.id"],
            name=op.f("fk_optimization_optimizationvariable_docs_dimension__id_ba30"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_optimizationvariable_docs")
        ),
        sa.UniqueConstraint(
            "dimension__id",
            name=op.f("uq_optimization_optimizationvariable_docs_dimension__id"),
        ),
    )
    op.create_table(
        "optimization_table_docs",
        sa.Column("description", sa.TEXT(), nullable=False),
        sa.Column("dimension__id", sa.INTEGER(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_table.id"],
            name=op.f("fk_optimization_table_docs_dimension__id_optimization_table"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_table_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_table_docs_dimension__id")
        ),
    )
    op.create_table(
        "optimization_variableindexsetassociation",
        sa.Column("variable__id", sa.INTEGER(), nullable=False),
        sa.Column("indexset__id", sa.INTEGER(), nullable=False),
        sa.Column("column_name", sa.VARCHAR(length=255), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_variableindexsetassociation_indexset__i_044c"),
        ),
        sa.ForeignKeyConstraint(
            ["variable__id"],
            ["optimization_optimizationvariable.id"],
            name=op.f("fk_optimization_variableindexsetassociation_variable__i_02a8"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_variableindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_variableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_variableindexsetassociation_variable__id"),
            ["variable__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_variableindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )

    op.create_table(
        "optimization_equation_docs",
        sa.Column("description", sa.TEXT(), nullable=False),
        sa.Column("dimension__id", sa.INTEGER(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_equation.id"],
            name=op.f("fk_optimization_equation_docs_dimension__id_optimizatio_bb11"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_equation_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_equation_docs_dimension__id")
        ),
    )

    op.create_table(
        "optimization_parameter_docs",
        sa.Column("description", sa.TEXT(), nullable=False),
        sa.Column("dimension__id", sa.INTEGER(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_parameter.id"],
            name=op.f("fk_optimization_parameter_docs_dimension__id_optimizati_898e"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_parameter_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_parameter_docs_dimension__id")
        ),
    )
    op.create_table(
        "optimization_tableindexsetassociation",
        sa.Column("table__id", sa.INTEGER(), nullable=False),
        sa.Column("indexset__id", sa.INTEGER(), nullable=False),
        sa.Column("column_name", sa.VARCHAR(length=255), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_tableindexsetassociation_indexset__id_o_868a"),
        ),
        sa.ForeignKeyConstraint(
            ["table__id"],
            ["optimization_table.id"],
            name=op.f("fk_optimization_tableindexsetassociation_table__id_opti_5cfe"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_tableindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_tableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_tableindexsetassociation_table__id"),
            ["table__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_tableindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )

    with op.batch_alter_table("opt_var_idx_association", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_var_idx_association_variable__id"))
        batch_op.drop_index(batch_op.f("ix_opt_var_idx_association_indexset__id"))

    op.drop_table("opt_var_idx_association")
    op.drop_table("opt_var_docs")
    with op.batch_alter_table("opt_tab_idx_association", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_tab_idx_association_table__id"))
        batch_op.drop_index(batch_op.f("ix_opt_tab_idx_association_indexset__id"))

    op.drop_table("opt_tab_idx_association")
    op.drop_table("opt_tab_docs")
    op.drop_table("opt_sca_docs")
    with op.batch_alter_table("opt_par_idx_association", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_par_idx_association_parameter__id"))
        batch_op.drop_index(batch_op.f("ix_opt_par_idx_association_indexset__id"))

    op.drop_table("opt_par_idx_association")
    op.drop_table("opt_par_docs")
    op.drop_table("opt_idx_docs")
    with op.batch_alter_table("opt_idx_data", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_idx_data_indexset__id"))

    op.drop_table("opt_idx_data")
    with op.batch_alter_table("opt_equ_idx_association", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_equ_idx_association_indexset__id"))
        batch_op.drop_index(batch_op.f("ix_opt_equ_idx_association_equation__id"))

    op.drop_table("opt_equ_idx_association")
    op.drop_table("opt_equ_docs")
    with op.batch_alter_table("opt_var_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_var_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_opt_var_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_opt_var_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_opt_var_version_end_transaction_id"))

    op.drop_table("opt_var_version")
    with op.batch_alter_table(
        "opt_var_idx_association_version", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_opt_var_idx_association_version_variable__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_var_idx_association_version_transaction_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_var_idx_association_version_operation_type")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_var_idx_association_version_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_var_idx_association_version_end_transaction_id")
        )

    op.drop_table("opt_var_idx_association_version")
    with op.batch_alter_table("opt_tab_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_tab_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_opt_tab_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_opt_tab_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_opt_tab_version_end_transaction_id"))

    op.drop_table("opt_tab_version")
    with op.batch_alter_table(
        "opt_tab_idx_association_version", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_opt_tab_idx_association_version_transaction_id")
        )
        batch_op.drop_index(batch_op.f("ix_opt_tab_idx_association_version_table__id"))
        batch_op.drop_index(
            batch_op.f("ix_opt_tab_idx_association_version_operation_type")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_tab_idx_association_version_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_tab_idx_association_version_end_transaction_id")
        )

    op.drop_table("opt_tab_idx_association_version")
    with op.batch_alter_table("opt_sca_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_sca_version_unit__id"))
        batch_op.drop_index(batch_op.f("ix_opt_sca_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_opt_sca_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_opt_sca_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_opt_sca_version_end_transaction_id"))

    op.drop_table("opt_sca_version")
    with op.batch_alter_table("opt_par_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_par_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_opt_par_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_opt_par_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_opt_par_version_end_transaction_id"))

    op.drop_table("opt_par_version")
    with op.batch_alter_table(
        "opt_par_idx_association_version", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_opt_par_idx_association_version_transaction_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_par_idx_association_version_parameter__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_par_idx_association_version_operation_type")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_par_idx_association_version_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_par_idx_association_version_end_transaction_id")
        )

    op.drop_table("opt_par_idx_association_version")
    with op.batch_alter_table("opt_idx_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_idx_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_opt_idx_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_opt_idx_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_opt_idx_version_end_transaction_id"))

    op.drop_table("opt_idx_version")
    with op.batch_alter_table("opt_idx_data_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_idx_data_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_opt_idx_data_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_opt_idx_data_version_indexset__id"))
        batch_op.drop_index(batch_op.f("ix_opt_idx_data_version_end_transaction_id"))

    op.drop_table("opt_idx_data_version")
    with op.batch_alter_table("opt_equ_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_equ_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_opt_equ_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_opt_equ_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_opt_equ_version_end_transaction_id"))

    op.drop_table("opt_equ_version")
    with op.batch_alter_table(
        "opt_equ_idx_association_version", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_opt_equ_idx_association_version_transaction_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_equ_idx_association_version_operation_type")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_equ_idx_association_version_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_equ_idx_association_version_equation__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_opt_equ_idx_association_version_end_transaction_id")
        )

    op.drop_table("opt_equ_idx_association_version")
    with op.batch_alter_table("opt_var", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_var_run__id"))

    op.drop_table("opt_var")
    with op.batch_alter_table("opt_tab", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_tab_run__id"))

    op.drop_table("opt_tab")
    with op.batch_alter_table("opt_sca", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_sca_unit__id"))
        batch_op.drop_index(batch_op.f("ix_opt_sca_run__id"))

    op.drop_table("opt_sca")
    with op.batch_alter_table("opt_par", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_par_run__id"))

    op.drop_table("opt_par")
    with op.batch_alter_table("opt_equ", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_equ_run__id"))

    op.drop_table("opt_equ")
    with op.batch_alter_table("opt_idx", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_opt_idx_run__id"))

    op.drop_table("opt_idx")
    # ### end Alembic commands ###



from typing import Any

from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Scalar

ScalarDocs = docs_model(Scalar)


class ScalarDocsRepository(BaseDocsRepository[Any]):
    model_class = ScalarDocs
    dimension_model_class = Scalar



import logging
import time
from collections.abc import Awaitable, Callable

from fastapi import Request, Response
from starlette.middleware.base import BaseHTTPMiddleware

logger = logging.getLogger(__name__)


class RequestTimeLoggerMiddleware(BaseHTTPMiddleware):
    async def dispatch(
        self, request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> Response:
        start_time = time.time()
        response = await call_next(request)
        process_time = time.time() - start_time
        logger.debug(f"Request process time: {process_time} seconds.")
        return response


class RequestSizeLoggerMiddleware(BaseHTTPMiddleware):
    async def dispatch(
        self, request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> Response:
        body = await request.body()
        logger.debug(f"Request body size: {len(body)} bytes.")
        return await call_next(request)



from typing import TYPE_CHECKING

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.db.filters import BaseFilter

from .. import base, versions
from .docs import UnitDocsRepository
from .model import Unit, UnitVersion

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend


class UnitVersionRepository(versions.VersionRepository[UnitVersion]):
    model_class = UnitVersion


class EnumerateKwargs(abstract.annotations.HasNameFilter, total=False):
    _filter: BaseFilter


class CreateKwargs(TypedDict, total=False):
    name: str


class UnitRepository(
    base.Creator[Unit],
    base.Deleter[Unit],
    base.Retriever[Unit],
    base.Enumerator[Unit],
    base.BulkDeleter[Unit],
    base.BulkUpserter[Unit],
    abstract.UnitRepository,
):
    docs: UnitDocsRepository
    versions: UnitVersionRepository
    model_class = Unit

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = UnitDocsRepository(*args)
        self.versions = UnitVersionRepository(*args)

        from .filter import UnitFilter

        self.filter_class = UnitFilter

    def add(self, name: str) -> Unit:
        unit = Unit(name=name)
        self.session.add(unit)
        return unit

    @guard("manage")
    def create(self, /, *args: str, **kwargs: Unpack[CreateKwargs]) -> Unit:
        return super().create(*args, **kwargs)

    @guard("manage")
    def delete(self, /, *args: Unpack[tuple[int]]) -> None:
        return super().delete(*args)

    @guard("view")
    def get(self, name: str) -> Unit:
        exc = db.select(Unit).where(Unit.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Unit.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Unit:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Unit.NotFound(id=id)

        return obj

    @guard("view")
    def list(self, /, **kwargs: Unpack[EnumerateKwargs]) -> list[Unit]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)



from typing import Any, ClassVar, cast

import pandas as pd

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from . import base


class Checkpoint(base.BaseModel):
    NotFound: ClassVar = abstract.Unit.NotFound
    NotUnique: ClassVar = abstract.Unit.NotUnique
    DeletionPrevented: ClassVar = abstract.Unit.DeletionPrevented

    id: int
    run__id: int
    transaction__id: int | None
    message: str


class CheckpointRepository(
    base.Creator[Checkpoint],
    base.Deleter[Checkpoint],
    base.Enumerator[Checkpoint],
    abstract.CheckpointRepository,
):
    model_class = Checkpoint
    prefix = "checkpoints/"

    def create(self, run__id: int, message: str) -> Checkpoint:
        return super().create(run__id=run__id, message=message)

    def delete(self, id: int) -> None:
        super().delete(id)

    def get_by_id(self, id: int) -> Checkpoint:
        res = self._get_by_id(id)
        return Checkpoint(**res)

    def enumerate(
        self, **kwargs: Unpack[abstract.checkpoint.EnumerateKwargs]
    ) -> list[Checkpoint] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(
        self, **kwargs: Unpack[abstract.checkpoint.EnumerateKwargs]
    ) -> list[Checkpoint]:
        json = cast(dict[str, Any], kwargs)
        return super()._list(json=json)

    def tabulate(
        self, **kwargs: Unpack[abstract.checkpoint.EnumerateKwargs]
    ) -> pd.DataFrame:
        json = cast(dict[str, Any], kwargs)
        return super()._tabulate(json=json)



from typing import Any

import numpy as np
import pandas as pd


def map_existing(
    df: pd.DataFrame,
    existing_df: pd.DataFrame,
    join_on: tuple[str, str],
    map: tuple[str, str],
    suffixes: tuple[str | None, str | None] = ("_x", "_y"),
) -> tuple[pd.DataFrame, np.ndarray[Any, np.dtype[np.str_]]]:
    _, join_to = join_on
    _, map_to = map
    existing_df = existing_df.rename(columns=dict([join_on, map]))[[join_to, map_to]]
    df = df.merge(existing_df, how="left", on=[join_to], suffixes=suffixes)
    missing = df.where(pd.isnull(df[map_to]))[join_to].dropna().unique()
    return df, missing



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types
from ixmp4.data.db import mixins

from .. import base, versions


class Unit(base.BaseModel, mixins.HasCreationInfo):
    __tablename__ = "unit"
    NotFound: ClassVar = abstract.Unit.NotFound
    NotUnique: ClassVar = abstract.Unit.NotUnique
    DeletionPrevented: ClassVar = abstract.Unit.DeletionPrevented

    # NOTE With only types.UniqueName, the non-unique base version of name is used
    name: types.UniqueName = db.Column(db.String(255), nullable=False, unique=True)


class UnitVersion(versions.DefaultVersionModel):
    __tablename__ = "unit_version"
    name: types.String = db.Column(db.String(255), nullable=False)

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username


version_triggers = versions.PostgresVersionTriggers(
    Unit.__table__, UnitVersion.__table__
)



from typing import TYPE_CHECKING, Any, TypeVar, Union

import pandas as pd

from ixmp4 import db
from ixmp4.db.utils.revert import select_for_id_map

from . import base

if TYPE_CHECKING:
    from .equation.model import Equation
    from .equation.repository import EquationRepository
    from .indexset.model import IndexSet
    from .indexset.repository import IndexSetRepository
    from .parameter.model import Parameter
    from .parameter.repository import ParameterRepository
    from .table.model import Table
    from .table.repository import TableRepository
    from .variable.model import OptimizationVariable
    from .variable.repository import VariableRepository


def validate_data(
    host: base.BaseModel,
    data: dict[str, Any],
    indexsets: list["IndexSet"],
    column_names: list[str] | None = None,
    has_extra_columns: bool = True,
) -> None:
    data_frame = pd.DataFrame.from_dict(data)

    # Can't validate ("values","units") or ("levels", "marginals") when they are present
    number_columns = (
        len(data_frame.columns) - 2 if has_extra_columns else len(data_frame.columns)
    )
    columns = (
        column_names if column_names else [indexset.name for indexset in indexsets]
    )

    # TODO for all of the following, we might want to create unique exceptions
    # Could me make both more specific by specifiying missing/extra columns?
    if number_columns < len(indexsets):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            f"Data is missing for some columns! \n Data: {data} \n "
            f"Columns: {columns}"
        )
    elif number_columns > len(indexsets):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            f"Trying to add data to unknown columns! \n Data: {data} \n "
            f"Columns: {columns}"
        )

    # We could make this more specific maybe by pointing to the missing values
    if data_frame.isna().any(axis=None):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            "The data is missing values, please make sure it "
            "does not contain None or NaN, either!"
        )

    limited_to_indexsets = {
        columns[i]: indexsets[i].data for i in range(len(indexsets))
    }

    # We can make this more specific e.g. highlighting all duplicate rows via
    # pd.DataFrame.duplicated(keep="False")
    if data_frame[limited_to_indexsets.keys()].value_counts().max() > 1:
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \nThe data contains duplicate rows!"
        )

    # Can we make this more specific? Iterating over columns; if any is False,
    # return its name or something?
    if (
        not data_frame[limited_to_indexsets.keys()]
        .isin(limited_to_indexsets)
        .all(axis=None)
    ):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            "The data contains values that are not allowed as per the IndexSets "
            "it is constrained to!"
        )


# NOTE | does not seem to work yet for stringified type hints
def _find_columns_linked_to_indexset(
    item: Union["Table", "Parameter", "Equation", "OptimizationVariable"], name: str
) -> list[str]:
    """Determine columns in `item`.data that are linked to IndexSet `name`.

    Only works for `items` linked to IndexSet `name`.
    """
    if not item.column_names:
        # The item's indexset_names must be a unique list of names
        return [name]
    else:
        # If we have column_names, we must also have indexsets
        assert item.indexset_names

        # Handle possible duplicate values
        return [
            item.column_names[i]
            for i in range(len(item.column_names))
            if item.indexset_names[i] == name
        ]


ReposForIdMapType = TypeVar(
    "ReposForIdMapType",
    "EquationRepository",
    "IndexSetRepository",
    "ParameterRepository",
    "TableRepository",
    "VariableRepository",
)


def create_id_map_subquery(
    transaction__id: int, run__id: int, repo: ReposForIdMapType
) -> db.sql.Subquery:
    old_items_subquery = select_for_id_map(
        model_class=repo.versions.model_class,
        run__id=run__id,
        transaction__id=transaction__id,
    ).subquery()
    new_items_subquery = select_for_id_map(
        model_class=repo.model_class, run__id=run__id
    ).subquery()

    return db.utils.create_id_map_subquery(
        old_exc=old_items_subquery, new_exc=new_items_subquery
    )



from datetime import datetime
from typing import TYPE_CHECKING, ClassVar, Protocol

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.core.exceptions import (
    IxmpError,
    NoDefaultRunVersion,
    RunIsLocked,
    RunLockRequired,
)

from . import base
from .annotations import HasRunFilter, IamcRunFilter

if TYPE_CHECKING:
    from . import Model, Scenario


class Run(base.BaseModel, Protocol):
    """Model run data model."""

    NoDefaultVersion: ClassVar[type[IxmpError]] = NoDefaultRunVersion
    IsLocked: ClassVar[type[IxmpError]] = RunIsLocked
    LockRequired: ClassVar[type[IxmpError]] = RunLockRequired

    model__id: int
    "Foreign unique integer id of the model."
    model: "Model"
    "Associated model."

    scenario__id: int
    "Foreign unique integer id of the scenario."
    scenario: "Scenario"
    "Associated scenario."

    version: int
    "Run version."
    is_default: bool
    "`True` if this is the default run version."

    created_at: datetime
    created_by: str

    updated_at: datetime
    updated_by: str

    lock_transaction: int | None

    def __str__(self) -> str:
        return f"<Run {self.id} model={self.model.name} \
            scenario={self.scenario.name} version={self.version} \
            is_default={self.is_default}>"


class EnumerateKwargs(HasRunFilter, total=False):
    iamc: IamcRunFilter | bool | None


class RunRepository(
    base.Creator,
    base.Deleter,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    def create(self, model_name: str, scenario_name: str) -> Run:
        """Creates a run with an incremented version number or version=1 if no versions
        exist. Will automatically create the models and scenarios if they don't exist
        yet.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The created run.
        """
        ...

    def delete(self, id: int) -> None:
        """Deletes a run and **all associated iamc, optimization and meta data**.

        Parameters
        ----------
        id : int
            The unique integer id of the run.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If the run with `id` does not exist.
        """
        ...

    def get(self, model_name: str, scenario_name: str, version: int) -> Run:
        """Retrieves a run.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.
        version : int
            The version number of this run.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If the run with `model_name`, `scenario_name` and `version` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The retrieved run.
        """
        ...

    def get_or_create(self, model_name: str, scenario_name: str) -> Run:
        """Tries to retrieve a run's default version
        and creates it if it was not found.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The retrieved or created run.
        """
        try:
            return self.get_default_version(model_name, scenario_name)
        except Run.NoDefaultVersion:
            return self.create(model_name, scenario_name)

    def get_default_version(self, model_name: str, scenario_name: str) -> Run:
        """Retrieves a run's default version.

        Parameters
        ----------
        model_name : str
            The name of a model.
        scenario_name : str
            The name of a scenario.

        Raises
        ------
        :class:`ixmp4.core.exceptions.NoDefaultRunVersion`:
            If no runs with `model_name`, `scenario_name` and `is_default=True` exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The retrieved run.
        """
        ...

    def get_by_id(self, id: int) -> Run:
        """Retrieves a Run by its id.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`.
            If the Run with `id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The retrieved Run.
        """
        ...

    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Run]:
        r"""Lists runs by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.run.filter.RunFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Run`]:
            List of runs.
        """
        ...

    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        r"""Tabulate runs by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.run.filter.RunFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - model__id
                - scenario__id
        """
        ...

    def set_as_default_version(self, id: int) -> None:
        """Sets a run as the default version for a (model, scenario) combination.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If no run with the `id` exists.

        """
        ...

    def unset_as_default_version(self, id: int) -> None:
        """Unsets a run as the default version leaving no
        default version for a (model, scenario) combination.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If no run with the `id` exists.
        :class:`ixmp4.core.exceptions.IxmpError`:
            If the run is not set as a default version.

        """
        ...

    def revert(
        self, id: int, transaction__id: int, revert_platform: bool = False
    ) -> None:
        """Reverts run data to a specific `transaction__id`.

        Parameters
        ----------
        id : int
            Unique integer id.
        transaction__id : int
            Id of the transaction to revert to.
        revert_platform : bool, optional
            Whether to revert the units defined on the platform, too. Default `False`.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If no run with the `id` exists.
        """
        ...

    def lock(self, id: int) -> Run:
        """Locks a run at the current transaction (via `transaction__id`).

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If no run with the `id` exists.
        :class:`ixmp4.data.abstract.Run.IsLocked`:
            If the run is already locked.

        """
        ...

    def unlock(self, id: int) -> Run:
        """Locks a run at the current transaction (via `transaction__id`).

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Run.NotFound`:
            If no run with the `id` exists.
        """
        ...

    def clone(
        self,
        run_id: int,
        model_name: str | None = None,
        scenario_name: str | None = None,
        keep_solution: bool = True,
    ) -> Run:
        """Clone all data from one run to a new one.

        Parameters
        ----------
        run_id: int
            The unique integer id of the base run.
        model_name: str | None
            The new name of the model used in the new run, optional.
        scenario_name: str | None
            The new name of the scenario used in the new run, optional.
        keep_solution: bool
            Whether to keep the solution data from the base run. Optional, defaults to
            `True`.

        Returns
        -------
        :class:`ixmp4.data.abstract.Run`:
            The clone of the base run.
        """
        ...



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.db import versions
from ixmp4.data.db.unit import Unit

from .. import base


class Scalar(base.BaseModel):
    __tablename__ = "opt_sca"

    NotFound: ClassVar = abstract.Scalar.NotFound
    NotUnique: ClassVar = abstract.Scalar.NotUnique
    DeletionPrevented: ClassVar = abstract.Scalar.DeletionPrevented

    run__id: types.RunId

    value: types.Float = db.Column(db.Float, nullable=True, unique=False)

    unit: types.Mapped[Unit] = db.relationship()
    unit__id: types.Mapped[int] = db.Column(
        db.Integer, db.ForeignKey("unit.id"), index=True
    )

    updateable_columns = ["value", "unit__id"]

    __table_args__ = (db.UniqueConstraint("name", "run__id"),)


class ScalarVersion(versions.DefaultVersionModel):
    __tablename__ = "opt_sca_version"

    name: types.String = db.Column(db.String(255), nullable=False)
    run__id: db.MappedColumn[int] = db.Column(db.Integer, nullable=False, index=True)
    value = db.Column(db.Float, nullable=True, unique=False)
    unit__id: db.MappedColumn[int] = db.Column(db.Integer, index=True)

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username


version_triggers = versions.PostgresVersionTriggers(
    Scalar.__table__, ScalarVersion.__table__
)



from collections.abc import Iterable
from typing import TYPE_CHECKING, Generic, Protocol, TypeVar

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4.data.abstract.unit import Unit

from .. import base
from ..docs import DocsRepository

if TYPE_CHECKING:
    from . import EnumerateKwargs

BackendModelType = TypeVar("BackendModelType", bound=base.BaseModel, covariant=True)


class CreateKwargs(TypedDict, total=False):
    value: float
    unit: str | Unit | None
    constrained_to_indexsets: str | list[str] | None
    column_names: list[str] | None


class BackendBaseRepository(
    Generic[BackendModelType],
    base.Creator,
    base.Deleter,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(
        self, run_id: int, name: str, **kwargs: Unpack["CreateKwargs"]
    ) -> BackendModelType: ...

    def delete(self, id: int) -> None: ...

    def get(self, run_id: int, name: str) -> BackendModelType: ...

    def list(
        self, **kwargs: Unpack["EnumerateKwargs"]
    ) -> Iterable[BackendModelType]: ...

    def tabulate(self, **kwargs: Unpack["EnumerateKwargs"]) -> pd.DataFrame: ...



from typing import ClassVar

from ixmp4.db import Session, filters, sql

from .. import Region, TimeSeries


class RegionFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)
    hierarchy: filters.String | None = filters.Field(None)

    sqla_model: ClassVar[type] = Region

    def join(
        self, exc: sql.Select[tuple[Region]], session: Session | None = None
    ) -> sql.Select[tuple[Region]]:
        exc = exc.join(Region, TimeSeries.region)
        return exc



from collections.abc import Iterable
from functools import partial, reduce
from typing import TYPE_CHECKING, Any, Literal

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.abstract.annotations import HasUnitIdFilter
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.data.db.unit.model import Unit, UnitVersion
from ixmp4.data.db.utils import map_existing
from ixmp4.db import utils
from ixmp4.db.utils.revert import apply_transaction__id, select_for_id_map

from .. import base
from .docs import ScalarDocsRepository
from .model import Scalar, ScalarVersion

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend
    from ixmp4.data.db.versions.model import DefaultVersionModel


class EnumerateKwargs(base.EnumerateKwargs, HasUnitIdFilter, total=False): ...


class ScalarVersionRepository(
    versions.VersionRepository[ScalarVersion], base.BulkUpserter[ScalarVersion]
):
    model_class = ScalarVersion

    def select(
        self,
        transaction__id: int | None = None,
        run__id: int | None = None,
        valid: Literal["at_transaction", "after_transaction"] = "at_transaction",
        revert_platform: bool = False,
        **kwargs: Any,
    ) -> db.sql.Select[Any]:
        exc = db.select(self.bundle, UnitVersion.name.label("unit")).select_from(
            self.model_class
        )

        exc = exc.join(UnitVersion, onclause=ScalarVersion.unit__id == UnitVersion.id)

        _apply_transaction__id = partial(
            apply_transaction__id, transaction__id=transaction__id, valid=valid
        )

        # NOTE If not reverting Units, don't limit the selection because of them
        models_to_consider: set[type["DefaultVersionModel"]] = {self.model_class}
        if revert_platform:
            models_to_consider.add(UnitVersion)

        exc = reduce(_apply_transaction__id, models_to_consider, exc)

        if run__id is not None:
            exc = exc.where(ScalarVersion.run__id == run__id)

        exc = utils.where_matches_kwargs(exc, model_class=self.model_class, **kwargs)
        return exc.distinct()


class ScalarRepository(
    base.Creator[Scalar],
    base.Deleter[Scalar],
    base.Retriever[Scalar],
    base.Enumerator[Scalar],
    base.Reverter[Scalar],
    abstract.ScalarRepository,
):
    model_class = Scalar
    versions: ScalarVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = ScalarDocsRepository(*args)

        from .filter import OptimizationScalarFilter

        self.filter_class = OptimizationScalarFilter

        self.versions = ScalarVersionRepository(*args)

    def add(self, name: str, value: float | int, unit_name: str, run_id: int) -> Scalar:
        unit_id = self.backend.units.get(unit_name).id
        scalar = Scalar(
            name=name,
            value=value,
            unit__id=unit_id,
            run__id=run_id,
        )
        self.session.add(scalar)
        return scalar

    @guard("edit")
    def delete(self, id: int) -> None:
        return super().delete(id=id)

    @guard("view")
    def get(self, run_id: int, name: str) -> Scalar:
        exc = db.select(Scalar).where(
            (Scalar.name == name) & (Scalar.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Scalar.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Scalar:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Scalar.NotFound(id=id)

        return obj

    @guard("edit")
    def create(self, name: str, value: float, unit_name: str, run_id: int) -> Scalar:
        return super().create(
            name=name, value=value, unit_name=unit_name, run_id=run_id
        )

    @guard("edit")
    def update(
        self, id: int, value: float | None = None, unit_id: int | None = None
    ) -> Scalar:
        scalar = self.get_by_id(id)

        if value is not None:
            scalar.value = value
        if unit_id is not None:
            scalar.unit__id = unit_id

        self.session.commit()

        return scalar

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> Iterable[Scalar]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    def _map_units_on_revert(self, df: pd.DataFrame) -> pd.DataFrame:
        """Maps foreign-key-unit-ids to the ids of restored Units."""
        # NOTE This df has two unit__id columns: one containing current ids and one ids
        # from before the rollback (called _old). Extra columns like this are ignored by
        # bulk_insert(), it seems.
        df, missing = map_existing(
            df,
            existing_df=self.backend.units.tabulate(name__in=df["unit"]),
            join_on=("name", "unit"),
            map=("id", "unit__id"),
            suffixes=("_old", None),
        )
        if len(missing) > 0:
            raise Unit.NotFound(", ".join(missing))

        return df

    @guard("edit")
    def revert(
        self, transaction__id: int, run__id: int, revert_platform: bool = False
    ) -> None:
        correct_versions: pd.DataFrame
        if revert_platform:
            old_unit_subquery = select_for_id_map(
                model_class=UnitVersion, run__id=None, transaction__id=transaction__id
            ).subquery()
            new_unit_subquery = select_for_id_map(
                model_class=Unit, run__id=None
            ).subquery()

            unit_map_subquery = utils.create_id_map_subquery(
                old_exc=old_unit_subquery, new_exc=new_unit_subquery
            )

            columns = utils.collect_columns_to_select(
                columns=utils.get_columns(self.versions.model_class),
                exclude={"unit__id"},
            )

            select_correct_versions = (
                db.select(
                    unit_map_subquery.c.new_id.label("unit__id"), *columns.values()
                )
                .select_from(self.versions.model_class)
                .join(
                    unit_map_subquery,
                    self.versions.model_class.unit__id == unit_map_subquery.c.old_id,
                )
            )

            select_correct_versions = apply_transaction__id(
                exc=select_correct_versions,
                model_class=self.versions.model_class,
                transaction__id=transaction__id,
            ).order_by(self.versions.model_class.id.asc())

            correct_versions = self.tabulate_query(select_correct_versions)
        else:
            correct_versions = self.versions.tabulate(
                transaction__id=transaction__id, revert_platform=revert_platform
            )

        return super().revert(
            transaction__id=transaction__id,
            run__id=run__id,
            correct_versions=correct_versions,
        )



from typing import TYPE_CHECKING

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from .. import versions

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend


from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.db.filters import BaseFilter

from .. import base
from .docs import ModelDocsRepository
from .model import Model, ModelVersion


class ModelVersionRepository(versions.VersionRepository[ModelVersion]):
    model_class = ModelVersion


class EnumerateKwargs(abstract.annotations.HasNameFilter, total=False):
    _filter: BaseFilter


class CreateKwargs(TypedDict, total=False):
    name: str


class ModelRepository(
    base.Creator[Model],
    base.Retriever[Model],
    base.Enumerator[Model],
    abstract.ModelRepository,
):
    model_class = Model
    docs: ModelDocsRepository
    versions: ModelVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = ModelDocsRepository(*args)
        self.versions = ModelVersionRepository(*args)

        from .filter import ModelFilter

        self.filter_class = ModelFilter

    def add(self, name: str) -> Model:
        model = Model(name=name)
        self.session.add(model)
        return model

    @guard("view")
    def get(self, name: str) -> Model:
        exc = self.select(name=name)
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Model.NotFound

    @guard("edit")
    def create(self, *args: str, **kwargs: Unpack[CreateKwargs]) -> Model:
        return super().create(*args, **kwargs)

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Model]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)



from collections.abc import Iterable
from typing import ClassVar

import pandas as pd

from ixmp4.data import abstract

from . import base


class Docs(base.BaseModel):
    NotFound: ClassVar = abstract.Docs.NotFound
    NotUnique: ClassVar = abstract.Docs.NotUnique
    DeletionPrevented: ClassVar = abstract.Docs.DeletionPrevented

    id: int
    description: str
    dimension__id: int


class DocsRepository(
    base.Creator[Docs],
    base.Retriever[Docs],
    base.Deleter[Docs],
    base.Enumerator[Docs],
    abstract.DocsRepository,
):
    model_class: type[Docs]
    enumeration_method = "GET"

    def get(self, dimension_id: int) -> Docs:
        return super().get(dimension_id=dimension_id)

    def set(self, dimension_id: int, description: str) -> Docs:
        res = self._create(
            self.prefix,
            json={
                "dimension_id": dimension_id,
                "description": description,
            },
        )
        return Docs(**res)

    # NOTE This is not used anywhere, but without it, mypy complains that the base
    # definitions of enumerate() are incompatible
    def enumerate(self, dimension_id: int | None = None) -> list[Docs] | pd.DataFrame:
        return super().enumerate(dimension_id=dimension_id)

    def list(
        self,
        *,
        dimension_id: int | None = None,
        dimension_id__in: Iterable[int] | None = None,
    ) -> list[Docs]:
        return super()._list(
            params={"dimension_id": dimension_id, "dimension_id__in": dimension_id__in}
        )

    def delete(self, dimension_id: int) -> None:
        super().delete(dimension_id)
